class InvalidServicePrefix(Exception):
    pass


class ModelNotFound(Exception):
    pass


class InvalidRuntimeException(Exception):
    pass


class UnsupportedModelFormat(Exception):
    pass


class NoCredentialsFound(Exception):
    pass


class InvalidCredentials(Exception):
    pass


class IntegrationError(Exception):
    pass


class InvalidModelURI(Exception):
    pass


class DatasetNotFound(Exception):
    pass
